/* Copyright (c) 2000 Antonios Christofides
 * 
 * dakt.c - most of Dakt is here
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include "texts.h"
#include "daktsys.h"

const char *Progname = "dakt";
const char *version = "2.0.2";
const char *releasedate = "Sep 2000";

static void dakt(void);
static void display_page (int, int, const char *);
static void read_section (enum sections);
static const char *do_exercise (EXERCISE *);
static EXERCISE *find_exercise (int, int);
static void showlicense(void);

int main(int argc, char *argv[])
{
	terminal_init();
	dakt();
	terminal_reset();
	exit(0);
}

/*---------------------------------------------------------------------------*/

static void dakt(void)
{
	int key_pressed=0;

	for (;;) {
		display_page (MAIN_MENU, 1, NULL);
		for (;;) {
			switch(key_pressed=read_key()) {
			case 'L': case 'l':
			  	showlicense();
				break;
			case 'I': case 'i':
				read_section(A);
				break;
			case '1': case '2': case '3': case '4': case '5':
			case '6':
				read_section(key_pressed-'1'+B);
				break;
			case 'T': case 't':
				read_section(H);
				break;
			case '7':
				read_section(I);
				break;
			case 'Q': case 'q':
				fputs("\n", stdout);
				return;
			case EOF:
				terminal_reset();
				fprintf(stderr, "%s: Error reading stdin: %s\n",
				Progname, ferror(stdin) ? strerror(errno) :
				"Unexpected end of file");
				exit(1);
			default:
				continue;
			}
			break;
		}
	}
}

/*---------------------------------------------------------------------------*/

static void showlicense(void)
{
	fputs("\n\n\n", stderr);
	fputs(License, stderr);
	fputs("\nPress any key to return to the main menu...", stderr);
	read_key();
}

/*---------------------------------------------------------------------------*/

static void display_page(int section, int current_page,
    					const char *exercise_result)
{
	static char buffer[81], bf[81];
	int line;
	EXERCISE *ex;

	fputs("\n\n", stdout);
	/* Read lines one by one, count, and display. */
	for (line = 1;;++line) {
		/* If at third line and there's an exercise in previous page,
		 * display exercise result.
		 */
		if(line==3 && (ex=find_exercise(section, current_page-1))!=NULL)
		{
			printf("%s\n\n", str_you_should_have_typed);
			fputs(ex->text, stdout);
			printf("\n\n%s\n\n", str_you_actually_typed);
			fputs(exercise_result==NULL?"":exercise_result, stdout);
			fputs("\n\n", stdout);
			line += 7;
			continue;
		}
		if(!get_text(section, current_page, line, buffer))
			break;

		/* If this is the first line of the main menu, print
		 * version and url.
		 */
		if(section==MAIN_MENU && line==1) {
			int i;
			sprintf(buffer, "%s%s (%s)", str_this_is_dakt_version,
			  version, releasedate);
			i = 79-strlen(url)-strlen(buffer);
			while(i--)
				strcat(buffer, " ");
			strcat(buffer, url);
		}
		/* Page number and title */
		if (section != MAIN_MENU  &&  line == 1) {
			char *p;

			for(p=strchr(buffer, '\0');p<buffer+80;++p)
			  	*p = '\0';
			strncpy(buffer, Title[section], strlen(Title[section]));
			sprintf(bf, "%-65s PAGE %2d OF %2d", buffer,
						current_page, Pages[section]);
			strcpy(buffer, bf);
		}

		fputs(buffer, stdout);
		if(line!=24 || section!=MAIN_MENU)
		  	putchar('\n');
	}

	while(section!=MAIN_MENU && line++<24)
		putchar('\n');
	if (section != MAIN_MENU)
		fputs(str_navigation_keys, stdout);
	fflush(stdout);
	if(ferror(stdout)) {
		terminal_reset();
		fprintf(stderr, "%s: Error writing to standard output: %s\n",
		Progname, strerror(errno));
		exit(1);
	}
}

/*----------------------------------------------------------------------------*/

static void read_section (enum sections section)
{
	int current_page = 0; /* Initialisation prevents compiler warning */
	EXERCISE *exercise;
	const char *exercise_result = "";
	enum { BEGIN, DISPLAY, DO_EXAMPLE, CHECK_EXERCISE_RESULT, PREVIOUS_PAGE,
		NEXT_PAGE, WAIT_FOR_KEY, DO_EXERCISE, END } state;

	state = BEGIN;
	for(;;)
		switch(state) {
		case BEGIN:
			current_page = 1;
			state = DISPLAY;
			break;
		case DISPLAY:
			display_page(section, current_page, exercise_result);
			state = (section==A && current_page==PAGE_EXAMPLE) ?
				DO_EXAMPLE : WAIT_FOR_KEY;
			break;
		case DO_EXAMPLE:
			exercise_result = do_exercise(NULL);
			state = CHECK_EXERCISE_RESULT;
			break;
		case CHECK_EXERCISE_RESULT:
			if(exercise_result==NULL)
				state = PREVIOUS_PAGE;
			else if(exercise_result==(const char *)0x1)
				state = END;
			else if(section==A && current_page==PAGE_EXAMPLE)
				state = NEXT_PAGE;
			else
				state = NEXT_PAGE;
			break;
		case PREVIOUS_PAGE:
			state = --current_page ? DISPLAY : END;
			break;
		case NEXT_PAGE:
			state = ++current_page>Pages[section] ? END : DISPLAY;
			break;
		case WAIT_FOR_KEY:
			switch(read_key()) {
				case '\n': state = NEXT_PAGE; break;
				case '\b': state = PREVIOUS_PAGE; break;
				case '\t': state = END; break;
				case ' ': state = DO_EXERCISE; break;
			}
			break;
		case DO_EXERCISE:
			if((exercise=find_exercise(section,current_page))==NULL)
				state = WAIT_FOR_KEY;
			else {
				exercise_result = do_exercise(exercise);
				state = CHECK_EXERCISE_RESULT;
			}
			break;
		case END:
			return;
		default:
			terminal_reset();
			fprintf(stderr, "%s: Internal program error in "
			"function read_section\n", Progname);
			exit(1);
		}
}

/*----------------------------------------------------------------------------*/

#define USER_INPUT_LENGTH 79
#define WAIT_TIME 800    /* ms between letter appear and disappear */

static const char *do_exercise (EXERCISE *exercise)
{
	static char user_input[USER_INPUT_LENGTH+1] = "";
	char *p = user_input;
	const char *next_letter;
	int key_pressed;

	fprintf(stderr, "\r%35s==>     <==                                 ",
	    	"");
	waitms(2000);
	for (next_letter = (exercise==NULL) ? "abcd " : exercise->text;
		exercise == NULL || *next_letter; ++next_letter)
	{
		/* If in example and string finished, redo from start */
		if (exercise==NULL && !*next_letter)
			next_letter = "abcd ";

		fprintf(stderr, "\r%35s==> '%c' <==", "", *next_letter);
		waitms(WAIT_TIME);
		fprintf(stderr, "\r%35s==>     <==", "");
		waitms(WAIT_TIME);

		/* Read keys. */
		while((key_pressed=get_key()))
                    switch (key_pressed) {
                         case '\n':
                              *p = '\0';
                              return user_input;
                         case '\b':
                              *p = '\0';
                              return NULL;
                         case '\t':
                              *p = '\0';
                              return (const char *) (0x1);
                         default:
                              if(key_pressed >= 32
                                 && key_pressed <=126
                                 && p<user_input + USER_INPUT_LENGTH)
                                   *(p++) = (char) key_pressed;
                              break;
                    }
	}
	*p = '\0';
	return user_input;
}

/*----------------------------------------------------------------------------*/

static EXERCISE *find_exercise(int section, int page)
{
	EXERCISE *p;

	for(p = Exercise; p->page > 0; ++p)
		if (p->section==section && p->page==page)
			return p;

	return NULL;
}
